# tape.rb -- simulated tape file format
#
# written 010809 by Mirian Crzig Lennox
# Copyright (c) 2001, released under the terms of the GNU Public License

require "cosmic"
require "pdp10"

module PDP10
  module Tape
    include PDP10

    attr_reader :record_start, :record_length, :record_end
    attr_reader :record_prevheader, :record_nextheader

    # low level reading/writing functions

    def read_int
      read(4).unpack1("V")
    end

    def write_int n
      write([n].pack("V"))
    end

    def write_pad
      write([0].pack("C"))
    end

    def write_mark
      write_int 0
    end

    def read_data len
      read(len)
    end

    def read_word
      splice((read_data(5) or return nil).unpack("NC"),32,4)
    end

    def write_word w
      write(shred(w,32,4).pack("NC"))
    end

    # methods to navigate the record structure of the tape format

    def set_record_info record_start, record_length
      @record_start = record_start
      @record_length = record_length
      @record_end = record_start + record_length
      @record_prevheader = record_start ? record_start - 4 : nil
      #$stdout.printf "record sync: start=%d, len=%d\n", record_start,
      #                record_length
      if (record_length == 0)
	@record_nextheader = nil
	return false
      end
      @record_nextheader = @record_end + (record_length % 2) + 4
      return true
    end

    def sync_record offset
      seek offset
      record_length = read_int
      record_start = tell
      return set_record_info record_start, record_length
    end

    def rewind_tape
      sync_record 0
    end

    def next_record
      if (@record_nextheader)
	return sync_record @record_nextheader 
      end
      return nil
    end

    def prev_record
      if (@record_prevheader)
	seek @record_prevheader
	return sync_record(tell + read_int)
      end
      return nil
    end

    def next_file
      return false if at_eot
      sync_record @record_nextheader until at_eof
      len = read_int
      if (len == 0)
	set_eot
	return false
      end
      set_record_info tell, len
      return true
    end

    # convenience methods to tell whether we're at a boundary condition

    def at_eof
      @record_nextheader == nil
    end

    def at_bof
      @record_prevheader == nil
    end

    def at_eot
      at_eof and at_bof
    end

    def set_eot
      @record_nextheader = nil
      @record_prevheader = nil
    end

    # read methods

    def read_data len
      unless ((remain = @record_end - tell) > 0)
	next_record
	return nil
      end
      if (remain < len)
	read(remain) + "\0" * (len - remain)
      else
	read(len)
      end
    end

    def read_record
      rec = []
      while (word = read_word)
	rec.push word
      end
      return rec
    end

    def read_file
      return nil if at_eot
      buf = []
      i = 0
      until at_eof
	while (word = read_word)
	  buf.push word
	  i += 1
	  #$stdout.printf "read file word %d\n", i
	end
      end
      next_file
      return buf
    end

    # write methods

    def write_record rec
      recbytes = rec.length * 5
      write_int recbytes
      rec.each do |word|
	write_word word
      end
      write_pad if recbytes.odd?
      write_int recbytes
    end

    def write_file buf
      words_remaining = buf.length
      idx = 0
      while words_remaining > 0
	recwords = words_remaining < 512 ? words_remaining : 512
	recbytes = recwords * 5
	#$stdout.printf "writing %d byte record\n", recbytes
	write_int recbytes
	for i in 1..recwords
	  if (word = buf[idx])
	    write_word word
	  else
	    fail "couldn't write word #{idx} of buf[#{buf.length}]\n"
	  end
	  idx += 1
	  words_remaining -= 1
	  #$stdout.printf "wrote word %d of %d\n", idx, buf.length
	end
	write_pad if recbytes.odd?
	write_int recbytes
      end
      write_mark
    end

  end
end
